#ifndef ELLIPTIC_ENERGY_HH
#define ELLIPTIC_ENERGY_HH

#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>
#include <dune/common/stdstreams.hh>

#include <dune/fufem/interval.hh>

#include "localnonlinearity.hh"

namespace Dune {
template <int dim> class EllipticEnergy {
public:
  typedef FieldVector<double, dim> SmallVector;
  typedef FieldMatrix<double, dim, dim> SmallMatrix;

  typedef LocalNonlinearity<dim> NonlinearityType;

  EllipticEnergy(SmallMatrix const &A, SmallVector const &b,
                 shared_ptr<NonlinearityType const> phi, int ignore = dim)
      : A(A), b(b), phi(phi), ignore(ignore) {}

  std::vector<double> const &get_kinks() const { return phi->get_kinks(); }

  double operator()(SmallVector const &v) const {
    SmallVector y(0);
    A.usmv(0.5, v, y);        //  1/2 Av
    y -= b;                   //  1/2 Av - b
    return y * v + (*phi)(v); // <1/2 Av - b,v> + H(|v|)
  }

  void descentAtZero(SmallVector &ret) const {
    SmallVector const zero(0);
    // If there is a direction of descent, this is it
    SmallVector d;
    smoothGradient(zero, d);
    d *= -1;

    Interval<double> D;
    phi->directionalSubDiff(zero, d, D);
    double const nonlinearDecline = D[1];
    double const smoothDecline = -(d * d);
    double const combinedDecline = smoothDecline + nonlinearDecline;

    if (combinedDecline < 0) {
      ret = d;
    } else {
      ret = 0;
    }
  }

  // returns false if the direction is tangential
  void descentDirectionNew(SmallVector const &x, SmallVector &ret) const {
    SmallVector pg;
    upperGradient(x, pg);
    SmallVector mg;
    lowerGradient(x, mg);
    double const pgx = pg * x;
    double const mgx = mg * x;
    if (pgx >= 0 && mgx >= 0) {
      ret = pg;
      ret *= -1;
    } else if (pgx <= 0 && mgx <= 0) {
      ret = mg;
      ret *= -1;
    } else {
      // We know that no descent can be expected in the tangential direction
      ret = SmallVector(0);
    }
  }

  bool descentDirection(SmallVector const &x, SmallVector &ret) const {
    // Check the squared norm rather than each component because
    // complementaryProjection() divides by it
    if (x.two_norm2() == 0.0) {
      descentAtZero(ret);
      return true;
    }

    SmallVector pg;
    upperGradient(x, pg);
    SmallVector mg;
    lowerGradient(x, mg);
    double const pgx = pg * x;
    double const mgx = mg * x;
    if (pgx >= 0 && mgx >= 0) {
      ret = pg;
      ret *= -1;
      return true;
    } else if (pgx <= 0 && mgx <= 0) {
      ret = mg;
      ret *= -1;
      return true;
    } else {
      // Includes the case that pg points in direction x and mg
      // points in direction -x. The projection will then be zero.
      SmallVector d;
      smoothGradient(x, d);
      complementaryProjection(d, x, ret);
      ret *= -1;
      return false;
    }
  }

  SmallMatrix const &A;
  SmallVector const &b;
  shared_ptr<NonlinearityType const> const phi;
  int const ignore; // Dimension that should be ignored; goes from 0
                    // to dim-1; the special value dim means that no
                    // dimension should be ignored

private:
  // Gradient of the smooth part
  void smoothGradient(SmallVector const &x, SmallVector &y) const {
    A.mv(x, y); // y = Av
    y -= b;     // y = Av - b
    if (ignore != dim)
      y[ignore] = 0;
  }

  void upperGradient(SmallVector const &x, SmallVector &y) const {
    phi->upperGradient(x, y);
    SmallVector z;
    smoothGradient(x, z);
    y += z;
    if (ignore != dim)
      y[ignore] = 0;
  }

  void lowerGradient(SmallVector const &x, SmallVector &y) const {
    phi->lowerGradient(x, y);
    SmallVector z;
    smoothGradient(x, z);
    y += z;
    if (ignore != dim)
      y[ignore] = 0;
  }

  // y = (id - P)(d) where P is the projection onto the line t*x
  void complementaryProjection(SmallVector const &d, SmallVector const &x,
                               SmallVector &y) const {
    double const dx = d * x;
    double const xx = x.two_norm2();
    y = d;
    y.axpy(-dx / xx, x);
  }
};
}
#endif
