#include <cassert>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/fufem/interval.hh>

#include <dune/tnnmg/problem-classes/bisection.hh>

#include <dune/tectonic/mydirectionalconvexfunction.hh>

#include "compute_state.hh"

// The nonlinearity exp(-x)
// NOTE: the Dieterich law has an additional linear term!
class DecayingExponential {
public:
  typedef Dune::FieldVector<double, 1> VectorType;
  typedef Dune::FieldMatrix<double, 1, 1> MatrixType;

  DecayingExponential(double tau) : tau(tau) {}

  void directionalSubDiff(VectorType const &u, VectorType const &v,
                          Interval<double> &D) const {
    D[0] = D[1] = v[0] * (-tau * std::exp(-u[0]));
  }

  void directionalDomain(VectorType const &, VectorType const &,
                         Interval<double> &dom) const {
    dom[0] = -std::numeric_limits<double>::max();
    dom[1] = std::numeric_limits<double>::max();
  }

private:
  double const tau;
};

double state_update_dieterich_bisection(double tau, double uol,
                                        double old_state) {
  DecayingExponential::VectorType const start(0);
  DecayingExponential::VectorType const direction(1);
  DecayingExponential const phi(tau);
  MyDirectionalConvexFunction<DecayingExponential> const J(
      1.0, old_state - uol, phi, start, direction);

  int bisectionsteps = 0;
  Bisection const bisection(0.0, 1.0, 1e-12, true, 0);    // TODO
  return bisection.minimize(J, 0.0, 0.0, bisectionsteps); // TODO
}

double state_update_dieterich(double tau, double uol, double old_state) {
  double const ret = state_update_dieterich_bisection(tau, uol, old_state);
  /* We have

     ret - old_state + uol = tau*exp(-ret)

     or

     log((ret - old_state + uol)/tau) = -ret
  */
  assert(std::min(std::abs(ret - old_state + uol - tau * std::exp(-ret)),
                  std::abs(std::log((ret - old_state + uol) / tau) + ret)) <
         1e-12);
  return ret;
}
