#ifndef TEST_GRADIENT_METHOD_NICEFUNCTION_HH
#define TEST_GRADIENT_METHOD_NICEFUNCTION_HH

#include <cassert>

#include <dune/tectonic/nicefunction.hh>

namespace Dune {
class MyFunction : public NiceFunction {
protected:
  MyFunction(std::vector<double> const &_kinks) : NiceFunction(_kinks) {}

public:
  MyFunction() : NiceFunction() {}

  double virtual second_deriv(double) const {
    assert(false);
    return 0;
  }

  double virtual regularity(double) const {
    assert(false);
    return 0;
  }
};

class Parabola : public MyFunction {
  double virtual evaluate(double x) const { return x * x; }

  double virtual leftDifferential(double s) const { return 2 * s; }

  double virtual rightDifferential(double s) const { return 2 * s; }

  double virtual second_deriv(double s) const { return 2; }

  double virtual regularity(double s) const { return 2; }

  bool virtual smoothesNorm() const { return true; }
};

class LinearFunction : public MyFunction {
public:
  LinearFunction(double a) : MyFunction(), coefficient(a) {}

  double virtual evaluate(double x) const { return coefficient * x; }

  double virtual leftDifferential(double s) const { return coefficient; }

  double virtual rightDifferential(double s) const { return coefficient; }

  double virtual second_deriv(double) const { return 0; }

  double virtual regularity(double s) const { return 0; }

private:
  double const coefficient;
};

template <int slope> class SampleFunction : public MyFunction {
public:
  SampleFunction() : MyFunction({ 1 }) {}

  double virtual evaluate(double x) const {
    return (x < 1) ? x : (slope * (x - 1) + 1);
  }

  double virtual leftDifferential(double s) const {
    return (s <= 1) ? 1 : slope;
  }

  double virtual rightDifferential(double s) const {
    return (s < 1) ? 1 : slope;
  }
};

// slope in [n-1,n] is n
class HorribleFunction : public MyFunction {
  // TODO: Handle kinks

public:
  double virtual evaluate(double x) const {
    double const fl = floor(x);
    double const sum = fl * (fl + 1) / 2;
    return sum + (fl + 1) * (x - fl);
  }

  double virtual leftDifferential(double x) const {
    double const fl = floor(x);
    if (x - fl < 1e-14)
      return fl;
    else
      return fl + 1;
  }

  double virtual rightDifferential(double x) const {
    double const c = ceil(x);
    if (c - x < 1e-14)
      return c + 1;
    else
      return c;
  }
};

// slope in [n-1,n] is log(n+1)
class HorribleFunctionLogarithmic : public MyFunction {
public:
  // TODO: Handle kinks

  double virtual evaluate(double x) const {
    double y = 0;
    size_t const fl = floor(x);
    for (size_t i = 1; i <= fl;)
      y += std::log(
          ++i); // factorials grow to fast so we compute this incrementally

    return y + std::log(fl + 2) * (x - fl);
  }

  double virtual leftDifferential(double x) const {
    double const fl = floor(x);
    if (x - fl < 1e-14)
      return std::log(fl + 1);
    else
      return std::log(fl + 2);
  }

  double virtual rightDifferential(double x) const {
    double const c = ceil(x);
    if (c - x < 1e-14)
      return std::log(c + 2);
    else
      return std::log(c + 1);
  }
};

class ThreeKinkFunction : public MyFunction {
public:
  ThreeKinkFunction() : MyFunction({ 5, 10, 15 }) {}

  double virtual evaluate(double x) const {
    double acc = 0;
    double last_kink = 0;
    int i;
    auto const &kinks = this->get_kinks();
    for (i = 0; i < kinks.size(); ++i) {
      if (x <= kinks[i])
        break;

      acc += (i + 1) * (kinks[i] - last_kink);
      last_kink = kinks[i];
    }
    return acc + (i + 1) * (x - last_kink);
  }

  double virtual leftDifferential(double x) const {
    if (x <= 5)
      return 1;
    if (x <= 10)
      return 2;
    if (x <= 15)
      return 3;
    return 4;
  }

  double virtual rightDifferential(double x) const {
    if (x < 5)
      return 1;
    if (x < 10)
      return 2;
    if (x < 15)
      return 3;
    return 4;
  }
};
}
#endif
