/* Checks if the descent direction is computed correctly using the
   analytic solution; also checks if the algorithm converges
   to the right solution regardless of where it starts */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <cassert>

#include <dune/common/shared_ptr.hh>

#include <dune/tectonic/ellipticenergy.hh>

#include "test-gradient-method-helper.hh"
#include "test-gradient-method-nicefunction.hh"

int main() {
  int const dim = 2;
  typedef Dune::EllipticEnergy<dim> Functional;
  typedef Functional::SmallMatrix SmallMatrix;
  typedef Functional::SmallVector SmallVector;

  SmallMatrix const A = { { 3, 1.5 }, { 1.5, 4 } };
  SmallVector const b = { 1, 2 };

  // |x|^2 as the nonlinearity is the same as having no nonlinearity
  // but twice the identity matrix added to A. In other words, we're
  // solving A + 2*id = b
  auto const f = Dune::make_shared<Dune::Parabola const>();
  auto const phi = Dune::make_shared<Functional::NonlinearityType const>(f);
  Functional const J(A, b, phi);

  SmallVector solution = { 4.0 / 37.0, 34.0 / 111.0 }; // Analytic solution
  SmallMatrix M = A;
  M[0][0] += 2;
  M[1][1] += 2;

  double ret1;
  {
    SmallVector start = b;
    start *= 17;

    SmallVector analytic_descent = b;
    M.mmv(start, analytic_descent);

    SmallVector numerical_descent;
    J.descentDirection(start, numerical_descent);
    assert(two_distance<dim>(numerical_descent, analytic_descent) < 1e-10);

    ret1 = functionTester(J, start, 6);
    assert(two_distance<dim>(start, solution) < 1e-6);
  }

  double ret2;
  {
    // Something random
    SmallVector start = { 279, -96 };
    SmallVector analytic_descent = b;
    M.mmv(start, analytic_descent);

    SmallVector numerical_descent;
    J.descentDirection(start, numerical_descent);
    assert(two_distance<dim>(numerical_descent, analytic_descent) < 1e-10);

    ret2 = functionTester(J, start, 15);
    assert(two_distance<dim>(start, solution) < 1e-6);
  }
  assert(std::abs(ret1 - ret2) < 1e-11);
}
